/***************************************************************************
                          MemoryOperationsHip.hpp  -  description
                             -------------------
    begin                : Jul 17, 2020
    copyright            : (C) 2020 by Tomas Oberhuber
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/* See Copyright Notice in tnl/Copyright */

#pragma once

#include <iostream>
#include <memory>  // std::unique_ptr
#include <stdexcept>

#include <TNL/Algorithms/MemoryOperations.h>
#include <TNL/Algorithms/MultiDeviceMemoryOperations.h>
#include <TNL/Algorithms/ParallelFor.h>
#include <TNL/Algorithms/Reduction.h>
#include <TNL/Exceptions/HipSupportMissing.h>

namespace TNL {
namespace Algorithms {

template< typename Element, typename Index >
void
MemoryOperations< Devices::Hip >::
construct( Element* data,
           const Index size )
{
   TNL_ASSERT_TRUE( data, "Attempted to create elements through a nullptr." );
   auto kernel = [data] __device_callable__ ( Index i )
   {
      // placement-new
      ::new( (void*) (data + i) ) Element();
   };
   ParallelFor< Devices::Hip >::exec( (Index) 0, size, kernel );
}

template< typename Element, typename Index, typename... Args >
void
MemoryOperations< Devices::Hip >::
construct( Element* data,
           const Index size,
           const Args&... args )
{
   TNL_ASSERT_TRUE( data, "Attempted to create elements through a nullptr." );
   // NOTE: nvcc does not allow __device_callable__ lambdas with a variadic capture
   auto kernel = [data] __device_callable__ ( Index i, Args... args )
   {
      // placement-new
      // (note that args are passed by value to the constructor, not via
      // std::forward or even by reference, since move-semantics does not apply for
      // the construction of multiple elements and pass-by-reference cannot be used
      // with GPU kernels)
      ::new( (void*) (data + i) ) Element( args... );
   };
   ParallelFor< Devices::Hip >::exec( (Index) 0, size, kernel, args... );
}

template< typename Element, typename Index >
void
MemoryOperations< Devices::Hip >::
destruct( Element* data,
          const Index size )
{
   TNL_ASSERT_TRUE( data, "Attempted to destroy data through a nullptr." );
   auto kernel = [data] __device_callable__ ( Index i )
   {
      (data + i)->~Element();
   };
   ParallelFor< Devices::Hip >::exec( (Index) 0, size, kernel );
}

template< typename Element >
__device_callable__ void
MemoryOperations< Devices::Hip >::
setElement( Element* data,
            const Element& value )
{
   TNL_ASSERT_TRUE( data, "Attempted to set data through a nullptr." );
#ifdef __HIP_DEVICE_COMPILE__
   *data = value;
#else
   // NOTE: calling `MemoryOperations< Devices::Hip >::set( data, value, 1 );`
   // does not work here due to `#ifdef __HIP_DEVICE_COMPILE__` above. It would
   // involve launching a GPU kernel with a __device_callable__ lambda, which
   // would be discarded by hipcc (it leads to a nice error "Cannot find Symbol").
   MultiDeviceMemoryOperations< Devices::Hip, void >::copy( data, &value, 1 );
#endif
}

template< typename Element >
__device_callable__ Element
MemoryOperations< Devices::Hip >::
getElement( const Element* data )
{
   TNL_ASSERT_TRUE( data, "Attempted to get data through a nullptr." );
#ifdef __HIP_DEVICE_COMPILE__
   return *data;
#else
   Element result;
   MultiDeviceMemoryOperations< void, Devices::Hip >::copy( &result, data, 1 );
   return result;
#endif
}

template< typename Element, typename Index >
void
MemoryOperations< Devices::Hip >::
set( Element* data,
     const Element& value,
     const Index size )
{
   if( size == 0 ) return;
   TNL_ASSERT_TRUE( data, "Attempted to set data through a nullptr." );
   auto kernel = [data, value] __device_callable__ ( Index i )
   {
      data[ i ] = value;
   };
   ParallelFor< Devices::Hip >::exec( (Index) 0, size, kernel );
}

template< typename DestinationElement,
          typename SourceElement,
          typename Index >
void
MemoryOperations< Devices::Hip >::
copy( DestinationElement* destination,
      const SourceElement* source,
      const Index size )
{
   if( size == 0 ) return;
   TNL_ASSERT_TRUE( destination, "Attempted to copy data to a nullptr." );
   TNL_ASSERT_TRUE( source, "Attempted to copy data from a nullptr." );

   // our ParallelFor kernel is faster than hipMemcpy
   auto kernel = [destination, source] __device_callable__ ( Index i )
   {
      destination[ i ] = source[ i ];
   };
   ParallelFor< Devices::Hip >::exec( (Index) 0, size, kernel );
}

template< typename DestinationElement,
          typename Index,
          typename SourceIterator >
void
MemoryOperations< Devices::Hip >::
copyFromIterator( DestinationElement* destination,
                  Index destinationSize,
                  SourceIterator first,
                  SourceIterator last )
{
   using BaseType = typename std::remove_cv< DestinationElement >::type;
   const int buffer_size = TNL::min( Hip::getTransferBufferSize() / sizeof(BaseType), destinationSize );
   std::unique_ptr< BaseType[] > buffer{ new BaseType[ buffer_size ] };
   Index copiedElements = 0;
   while( copiedElements < destinationSize && first != last ) {
      Index i = 0;
      while( i < buffer_size && first != last )
         buffer[ i++ ] = *first++;
      MultiDeviceMemoryOperations< Devices::Hip, void >::copy( &destination[ copiedElements ], buffer.get(), i );
      copiedElements += i;
   }
   if( first != last )
      throw std::length_error( "Source iterator is larger than the destination array." );
}

template< typename Element1,
          typename Element2,
          typename Index >
bool
MemoryOperations< Devices::Hip >::
compare( const Element1* destination,
         const Element2* source,
         const Index size )
{
   if( size == 0 ) return true;
   TNL_ASSERT_TRUE( destination, "Attempted to compare data through a nullptr." );
   TNL_ASSERT_TRUE( source, "Attempted to compare data through a nullptr." );

   auto fetch = [=] __device_callable__ ( Index i ) -> bool { return destination[ i ] == source[ i ]; };
   return Reduction< Devices::Hip >::reduce( ( Index ) 0, size, fetch, std::logical_and<>{}, true );
}

template< typename Element,
          typename Index >
bool
MemoryOperations< Devices::Hip >::
containsValue( const Element* data,
               const Index size,
               const Element& value )
{
   if( size == 0 ) return false;
   TNL_ASSERT_TRUE( data, "Attempted to check data through a nullptr." );
   TNL_ASSERT_GE( size, (Index) 0, "" );

   auto fetch = [=] __device_callable__ ( Index i ) -> bool { return data[ i ] == value; };
   return Reduction< Devices::Hip >::reduce( ( Index ) 0, size, fetch, std::logical_or<>{}, false );
}

template< typename Element,
          typename Index >
bool
MemoryOperations< Devices::Hip >::
containsOnlyValue( const Element* data,
                   const Index size,
                   const Element& value )
{
   if( size == 0 ) return false;
   TNL_ASSERT_TRUE( data, "Attempted to check data through a nullptr." );
   TNL_ASSERT_GE( size, 0, "" );

   auto fetch = [=] __device_callable__ ( Index i ) -> bool { return data[ i ] == value; };
   return Reduction< Devices::Hip >::reduce( ( Index ) 0, size, fetch, std::logical_and<>{}, true );
}

} // namespace Algorithms
} // namespace TNL
