/***************************************************************************
                          DeviceInfo.hpp  -  description
                             -------------------
    begin                : Jun 21, 2015
    copyright            : (C) 2007 by Tomas Oberhuber
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/* See Copyright Notice in tnl/Copyright */

#pragma once

#include <unordered_map>

#ifdef HAVE_HIP
#include <hip/hip_runtime.h>
#endif

#include <TNL/Hip/DeviceInfo.h>
#include <TNL/Exceptions/HipSupportMissing.h>

namespace TNL {
namespace Hip {

inline int
DeviceInfo::
getNumberOfDevices()
{
#ifdef HAVE_HIP
   int devices;
   hipGetDeviceCount( &devices );
   return devices;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getActiveDevice()
{
#ifdef HAVE_HIP
   int device;
   hipGetDevice( &device );
   return device;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline String
DeviceInfo::
getDeviceName( int deviceNum )
{
#ifdef HAVE_HIP
   hipDeviceProp_t properties;
   hipGetDeviceProperties( &properties, deviceNum );
   return String( properties.name );
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getArchitectureMajor( int deviceNum )
{
#ifdef HAVE_HIP
   hipDeviceProp_t properties;
   hipGetDeviceProperties( &properties, deviceNum );
   return properties.major;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getArchitectureMinor( int deviceNum )
{
#ifdef HAVE_HIP
   hipDeviceProp_t properties;
   hipGetDeviceProperties( &properties, deviceNum );
   return properties.minor;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getClockRate( int deviceNum )
{
#ifdef HAVE_HIP
   hipDeviceProp_t properties;
   hipGetDeviceProperties( &properties, deviceNum );
   return properties.clockRate;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline std::size_t
DeviceInfo::
getGlobalMemory( int deviceNum )
{
#ifdef HAVE_HIP
   hipDeviceProp_t properties;
   hipGetDeviceProperties( &properties, deviceNum );
   return properties.totalGlobalMem;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline std::size_t
DeviceInfo::
getFreeGlobalMemory()
{
#ifdef HAVE_HIP
   std::size_t free = 0;
   std::size_t total = 0;
   hipMemGetInfo( &free, &total );
   return free;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getMemoryClockRate( int deviceNum )
{
#ifdef HAVE_HIP
   hipDeviceProp_t properties;
   hipGetDeviceProperties( &properties, deviceNum );
   return properties.memoryClockRate;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline bool
DeviceInfo::
getECCEnabled( int deviceNum )
{
#ifdef HAVE_HIP
   hipDeviceProp_t properties;
   hipGetDeviceProperties( &properties, deviceNum );
   return properties.ECCEnabled;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getHipMultiprocessors( int deviceNum )
{
#ifdef HAVE_HIP
   // results are cached because they are used for configuration of some kernels
   static std::unordered_map< int, int > results;
   if( results.count( deviceNum ) == 0 ) {
      hipDeviceProp_t properties;
      hipGetDeviceProperties( &properties, deviceNum );
      results.emplace( deviceNum, properties.multiProcessorCount );
      return properties.multiProcessorCount;
   }
   return results[ deviceNum ];
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getHipCoresPerMultiprocessors( int deviceNum )
{
#ifdef HAVE_HIP
   // TODO: check if this is general enough
   // 64 taken from https://en.wikipedia.org/wiki/Graphics_Core_Next#Compute_units
   return 64;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getHipCores( int deviceNum )
{
#ifdef HAVE_HIP
   return DeviceInfo::getHipMultiprocessors( deviceNum ) *
          DeviceInfo::getHipCoresPerMultiprocessors( deviceNum );
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getRegistersPerMultiprocessor( int deviceNum )
{
   // TODO: regsPerMultiprocessor is not part of hipDeviceProp_t yet.
//   throw std::runtime_error( "HIP cannot detect number of registers per multiprocessor." );
   return 65536;
#ifdef HAVE_HIP
   // results are cached because they are used for configuration of some kernels
   static std::unordered_map< int, int > results;
   if( results.count( deviceNum ) == 0 ) {
      hipDeviceProp_t properties;
      hipGetDeviceProperties( &properties, deviceNum );
      //results.emplace( deviceNum, properties.regsPerMultiprocessor );
      //return properties.regsPerMultiprocessor;
   }
   return results[ deviceNum ];
#else
   throw Exceptions::HipSupportMissing();
#endif
}

} // namespace Hip
} // namespace TNL
